from typing import Dict, Optional, Protocol, TypeVar
import gymnasium
from pettingzoo import AECEnv


Action = TypeVar("Action")
Observation = TypeVar("Observation")


class ActOther(Protocol):
  def act(observation: Observation) -> Action:
    ...


class PZ2GymnasiumWrapper(gymnasium.Env):
  """
  This class wraps a PettingZoo environment (AECEnv) and exposes a simple Gymnasium environment (for single agent).
  In order to make it work, one needs to provide in the initialization a mechanism to get the actions of all other agents.
  """

  def __init__(self, pz_env: AECEnv, act_others: Dict[str, ActOther], take_spaces_from: str = None):
    super().__init__()
    self.pz_env = pz_env
    self.act_others = act_others
    take_spaces_from = take_spaces_from or next(iter(act_others)) # just use any of the keys in the act_others dict.
    self.observation_space = self.pz_env.observation_space(take_spaces_from)
    self.action_space = self.pz_env.action_space(take_spaces_from)

  def reset(
    self,
    seed: Optional[int] = None,
    options: Optional[dict] = None,
  ):
    super().reset(seed=seed)
    self.pz_env.reset(seed=seed)
    self._loop_others()
    agent = self.pz_env.agent_selection
    observation, info = (
      self.pz_env.observe(agent),
      self.pz_env.infos[agent]
    )
    return observation, info

  def step(self, action):
    agent = self.pz_env.agent_selection
    assert agent not in self.act_others, f"expected it to be my turn, got {agent}"
    self.pz_env.step(action)
    # agent = self.pz_env.agent_selection
    # print(f"agent now after step is {agent}")
    self._loop_others()
    observation, reward, terminated, truncated, info = (
      self.pz_env.observe(agent),
      self.pz_env.rewards[agent],
      self.pz_env.terminations[agent],
      self.pz_env.truncations[agent],
      self.pz_env.infos[agent]
    )
    # print(reward)
    return observation, reward, terminated, truncated, info

  def _loop_others(self):
    agent = self.pz_env.agent_selection
    # print(f"agent now is {agent}")
    while agent in self.act_others:
      act_current: ActOther = self.act_others[agent]
      obs_current: Observation = self.pz_env.observe(agent)
      action_current: Action = act_current(obs_current)
      self.pz_env.step(action_current)
      agent = self.pz_env.agent_selection

  def render(self, *args, **kwargs):
    self.pz_env.render(*args, **kwargs)



import gymnasium as gym
import random


class BJEnv(gym.Env):
    """
    Blackjack
    """
    
    class State:
        def __init__(self):
            # for the player I start with the sum rather than the cards to assure uniformity in the randomized start
            self.player_sum = random.randint(11, 21)
            # and so for the useful Ace
            self.player_useful_Ace = random.randint(0, 1) == 1 # boolean

            self.dealer_cards = [self._random_card(), self._random_card()]

        def _random_card(self):
            return random.randint(1, 10) # 1 - Ace, 2, .. 10

        def hits(self):
            new_card = self._random_card()
            self.player_sum += new_card # note since we start with 11 we cannot add another ACE.
            if self.player_sum > 21:
                if self.player_useful_Ace:
                    self.player_sum -= 10
                    self.player_useful_Ace = False
            return new_card
        
        def stick(self):                    
            return self._dealers_turn()
            
        def _dealers_turn(self):
            cur_sum = sum(self.dealer_cards)
            if cur_sum <= 11 and 1 in self.dealer_cards:
                cur_sum += 10
            while cur_sum < 17:
                self.dealer_cards.append(self._random_card())
                cur_sum = sum(self.dealer_cards)
                if cur_sum <= 11 and 1 in self.dealer_cards:
                    cur_sum += 10
            return cur_sum

        def render(self):
            if self.player_useful_Ace:
                print("Useful Ace")
            else:
                print("No useful Ace")
            print("Current sum: ", self.player_sum)
            print("Dealer showing: ",  self.card_to_image(self.dealer_cards[0]))
            
        def card_to_image(self, value):
            return "A" if value == 1 else value

                    
    def __init__(self):
        self.action_space = gym.spaces.Discrete(2) # 0 hits, 1 sticks
        
        obs_space = dict(
            player_sum = gym.spaces.Discrete(11), # x -> x + 11
            player_useful_Ace = gym.spaces.Discrete(2), # 0 no, 1 yes
            dealer_shown_card = gym.spaces.Discrete(10) # 0 - Ace, x -> x + 1 (ex. 1 is 2, 9 is 10)
        )
        
        self.observation_space = gym.spaces.Dict(obs_space)
        self.render_mode = None
        self.reset()
    def render(self):
        self.state.render()
    def reset(self, seed=None, options=None):
        if seed:
          random.seed(seed)
        self.state = BJEnv.State()
        return self._state_to_obs(), {}
    def _state_to_obs(self):
        return {
            'player_sum' : self.state.player_sum - 11,
            'player_useful_Ace': 1 if self.state.player_useful_Ace else 0,
            'dealer_shown_card' : self.state.dealer_cards[0] - 1
        }
    def step(self, action):
        done = None
        reward = None
        info = {}
        if action == 1: # sticks
            dealer_sum = self.state.stick()
            if dealer_sum > 21:
                reward = 1 # dealer busted
            elif self.state.player_sum < dealer_sum:
                reward = -1 # lose
            elif self.state.player_sum == dealer_sum:
                reward = 0 # draw
            else:
                reward = 1 # win
            info['dealer sum'] = dealer_sum 
            done = True
        elif action == 0: # hits
            new_card = self.state.hits()
            done = False
            reward = 0
            info['new card'] = self.state.card_to_image(new_card)
            if self.state.player_sum > 21:
                reward = -1
                done = True
        else:
            assert False, f"unkown action {action}"
        return self._state_to_obs(), reward, done, False, info


from qwertyenv.ensure_valid_action_pz import EnsureValidAction
from qwertyenv.pz_to_gymnasium_wrapper import PZ2GymnasiumWrapper

from tianshou.data import Collector
from tianshou.env import DummyVectorEnv
from tianshou.policy import RandomPolicy


from qwertyenv.collect_coins_pz import CollectCoinsEnv

from gymnasium.utils.env_checker import check_env


def test_evaluate():

  env = CollectCoinsEnv(pieces=['rock', 'rock']) # Note: PettingZoo environment (AECenv)

  action = None

  def another_action_taken(action_taken):
    nonlocal action
    action = action_taken

  # Wrapping the original environment as to make sure a valid action will be taken.
  env = EnsureValidAction(  # Still a PettingZoo environment
      env,
      env.check_action_valid,
      env.provide_alternative_valid_action,
      another_action_taken
  )

  def act_agent_1(obs):
      return (0, 0)

  env = PZ2GymnasiumWrapper(env, act_others={'agent_1': act_agent_1}) # Note: Gymnasium environment

  check_env(env) # , num_cycles=10, verbose_progress=False)


def test_with_tianshou():

  action = None

  # env =  gym.make('qwertyenv/CollectCoins-v0', pieces=['rock', 'rock'])

  env = CollectCoinsEnv(pieces=['rock', 'rock'], with_mask=True)  # Note: PettingZoo environment (AECenv)

  def another_action_taken(action_taken):
    nonlocal action
    action = action_taken

  # Wrapping the original environment as to make sure a valid action will be taken.
  env = EnsureValidAction(  # Still a PettingZoo environment
      env,
      env.check_action_valid,
      env.provide_alternative_valid_action,
      another_action_taken
  )

  def act_agent_1(obs):
      return (0, 0)

  env = PZ2GymnasiumWrapper(env, act_others={'agent_1': act_agent_1}) # Note: Gymnasium environment

  env = DummyVectorEnv([lambda: env])

  policy = RandomPolicy()

  collector = Collector(policy, env)

  result = collector.collect(n_step=200, render=0.1)



from gymnasium.envs.registration import register
from .ensure_valid_action import EnsureValidAction
from .up_down_left_right import UpDownLeftRight

__version__ = "0.1.0"


register(
    id='qwertyenv/BlackJack-v0',
    entry_point='qwertyenv.black_jack:BJEnv',
    max_episode_steps=300
)

register(
    id='qwertyenv/CollectCoins-v0',
    entry_point='qwertyenv.collect_coins:CollectCoinsEnv',
    max_episode_steps=300
)

__all__ = [
    EnsureValidAction,
    UpDownLeftRight
]


import pytest

import gym
from qwertyenv import EnsureValidAction
from stable_baselines3 import PPO
from stable_baselines3.ppo.policies import MultiInputPolicy
from stable_baselines3.common.evaluation import evaluate_policy


def test_evaluate():
  action = None

  env =  gym.make('qwertyenv/CollectCoins-v0', pieces=['rock', 'rock'])

  def another_action_taken(action_taken):
    nonlocal action
    action = action_taken

  # Wrapping the original environment as to make sure a valid action will be taken.
  env = EnsureValidAction(
      env,
      env.check_action_valid,
      env.provide_alternative_valid_action,
      another_action_taken
  )
  agent_w = PPO(MultiInputPolicy, env)
  for episode in range(1):
    # print(f'{episode=}')
    # print("-------")
    obs = env.reset()
    # env.render()
    while True:
      action, _state = agent_w.predict(obs)
      # print(f'action predicted: {action}')
      obs, reward, done, info = env.step(action)
      # print(f'action taken: {action}')
      # env.render()
      # print(f'{reward=}')
      # print(f'{done=}')
      # print(f'{info=}')
      if done:
        break
    # print()

  mean_reward, std_reward = evaluate_policy(agent_w, env, deterministic=True)



import functools
from itertools import product
import gymnasium as gym
import numpy as np
from pettingzoo import AECEnv
from pettingzoo.utils import agent_selector, wrappers
import random

from qwertyenv.collect_coins_game import CollectCoinsGame


class CollectCoinsEnv(AECEnv):
  """
  CollectCoins PettingZoo environment.

  This a Chess like, two players turns board game.
  The aim of the game is to collect more coins than the opponent.
  On each turn a the player's piece makes a move.
  If a coin is present in the destination the count is increased for the player.
  The game ends when there are not more coins on the board to collect. We then compare which player got the most.

  Note: a 'rock' can make a 1 distance step only. In the future we'll have also the regular interpretation for rock.

  Note: pieces don't eat one another and cannot step on each other.
  Pieces can only move to a free square with a legal move, eather collecting a coin, or not collecting a conin if there is no coin in the destination. 
  """

  metadata = {"render_modes": ["human"], "name": "CollectCoins_v0"}

  def __init__(self, pieces=['rock', 'rock'], render_mode=None, with_mask=False):
    """
    player 0 is the first entry, AKA white
    player 1 is the second entry, AKA black
    """

    self.render_mode = render_mode
    self.with_mask = with_mask
    self.possible_agents = ["player_" + str(r) for r in range(2)]
    self.agent_name_mapping = dict(
        zip(self.possible_agents, list(range(len(self.possible_agents))))
    )
    obs_space = dict(
      board = gym.spaces.MultiBinary([8, 8]),
      player = gym.spaces.Tuple([gym.spaces.Discrete(8)] * 2),
      other_player = gym.spaces.Tuple([gym.spaces.Discrete(8)] * 2)
    )    
    self._observation_space = gym.spaces.Dict(spaces=obs_space)
    self._action_space = gym.spaces.Tuple([gym.spaces.Discrete(8)] * 2)

    self.pieces = pieces
    self.game = None
    self.previous_coins = None
    self.reset()

  # this cache ensures that same space object is returned for the same agent
  # allows action space seeding to work as expected
  @functools.lru_cache(maxsize=None)
  def observation_space(self, agent):
      # gymnasium spaces are defined and documented here: https://gymnasium.farama.org/api/spaces/
      return self._observation_space

  @functools.lru_cache(maxsize=None)
  def action_space(self, agent):
      return self._action_space

  def reset(self, seed=None, options=None, return_info=False):
    """
    Reset needs to initialize the following attributes
    - agents
    - rewards
    - _cumulative_rewards
    - terminations
    - truncations
    - infos
    - agent_selection
    And must set up the environment so that render(), step(), and observe()
    can be called without issues.
    Here it sets up the state dictionary which is used by step() and the observations dictionary which is used by step() and observe()
    """
    self.game = CollectCoinsGame(self.pieces)

    self.agents = self.possible_agents[:]

    """
    Our agent_selector utility allows easy cyclic stepping through the agents list.
    """
    self._agent_selector = agent_selector(self.agents)
    self.agent_selection = self._agent_selector.next()

    self.rewards = {agent: 0 for agent in self.agents}
    self._cumulative_rewards = {agent: 0 for agent in self.agents}
    self.terminations = {agent: False for agent in self.agents}
    self.truncations = {agent: False for agent in self.agents}
    self.infos = {agent: {} for agent in self.agents}
    self.state = {agent: None for agent in self.agents}
    self.observations = {
      agent: self._get_observation(self.agent_name_mapping[agent]) for agent in self.agents
    }
    self.num_moves = 0

    self.previous_coins = [0, 0]

  def step(self, action):
      """
      step(action) takes in an action for the current agent (specified by
      agent_selection) and needs to update
      - rewards
      - _cumulative_rewards (accumulating the rewards)
      - terminations
      - truncations
      - infos
      - agent_selection (to the next agent)
      And any internal state used by observe() or render()
      """

      # print(f"{self.agent_selection=} {action=}")

      # print(type(action))

      if isinstance(action, np.int64): 
          action = (action // 8, action % 8)  # because for example, Tianshou may return np.int64 ?

      # action = tuple(action) # because for example, Tianshou may return list
      assert isinstance(action, tuple)
      assert len(action) == 2

      if (
          self.terminations[self.agent_selection]
          or self.truncations[self.agent_selection]
      ):
          # handles stepping an agent which is already dead
          # accepts a None action for the one agent, and moves the agent_selection to
          # the next dead agent,  or if there are no more dead agents, to the next live agent
          self._was_dead_step(action)
          return

      agent = self.agent_selection
      agent_idx = self.agent_name_mapping[agent]
      self.game.make_move(agent_idx, action)

      # the agent which stepped last had its _cumulative_rewards accounted for
      # (because it was returned by last()), so the _cumulative_rewards for this
      # agent should start again at 0
      self._cumulative_rewards[agent] = 0

      # stores action of current agent
      self.state[agent] = action

      # collect reward if it is the last agent to act
      if self._agent_selector.is_last():
          # rewards for all agents are placed in the .rewards dictionary
          self.rewards.update({
            self.agents[idx]: self._calc_reward(idx) for idx in range(2) 
          })

          done = self.game.is_done()
          if done:
            self.terminations.update({
              player: True
              for player in self.terminations
            })
          else:
            self.num_moves += 1
            # The truncations dictionary must be updated for all players.

            truncation = self.num_moves > 200 

            self.truncations.update({
               player: truncation
              for player in self.truncations
            })

          # observe the current state
          for i in self.agents:
            self.observations[i] = self._get_observation(self.agent_name_mapping[i]) # self.state[
            #     self.agents[1 - self.agent_name_mapping[i]]
            # ]
      else:
          # necessary so that observe() returns a reasonable observation at all times.
          self.state[self.agents[1 - agent_idx]] = 0 # TODO:??
          # no rewards are allocated until both players give an action
          self._clear_rewards()

      # selects the next agent.
      self.agent_selection = self._agent_selector.next()

      # print(f"{self.agent_selection=}")

      # Adds .rewards to ._cumulative_rewards
      self._accumulate_rewards()

      if self.render_mode == "human":
          self.render()

  def _get_observation(self, player_idx: int):
    obs = dict(
      board=self.game.board,
      player=self.game.locations[player_idx],
      other_player=self.game.locations[1 - player_idx]
    )

    if self.with_mask:
      all_moves = product(range(8), repeat=2)
      mask = [self.check_action_valid(move, player_idx) for move in all_moves]
      # mask = np.array([
      #    [self.check_action_valid((row, col), player) for col in range(8)]
      #    for row in range(8)
      # ])
      obs['mask'] = mask # TODO: should return a 8 x 8 matrix?

    return obs

  def _calc_reward(self, player_idx: int):
    current_coins = self.game.coins[player_idx]
    previous_coins = self.previous_coins[player_idx]
    self.previous_coins[player_idx] = current_coins
    done = self.game.is_done()
    if done:
      draw = self.game.coins[player_idx] == self.game.coins[1 - player_idx]
      win = self.game.coins[player_idx] > self.game.coins[1 - player_idx]
      return 0 if draw else (1 if win else -1)
    else:
      return (current_coins - previous_coins) * 0.01

  def render(self, *args, **argv):
    self.game.render()

  def observe(self, agent):
      """
      Observe should return the observation of the specified agent. This function
      should return a sane observation (though not necessarily the most up to date possible)
      at any time after reset() is called.
      """
      # observation of one agent is the previous state of the other
      return self.observations[agent]

  def close(self):
      """
      Close should release any graphical displays, subprocesses, network connections
      or any other environment data which should not be kept around after the
      user is no longer using the environment.
      """
      pass

  def check_action_valid(self, action, player_idx: int = None) -> bool:
    """
    This helper function can be used when initializing EnsureValidAction wrapper as an example (see in examples).
    """

    player_idx = player_idx or self.agent_name_mapping[self.agent_selection]

    # print(action)

    if not isinstance(action, (tuple, np.ndarray, list)):
        action = (action // 8, action % 8)  # because for example, Tianshou may return np.int64 ?

    # action = tuple(action)
    assert isinstance(action, tuple)
    assert len(action) == 2

    return self.game.valid_move(
      player_idx,
      action
    )

  def provide_alternative_valid_action(self, action, player: str=None):
    """
    This helper function can be used when initializing EnsureValidAction wrapper as an example (see in examples).

    Select a random valid move.
    TOOD: maybe consult the provided action(move) and provide an action that is as close as possible.
    """

    player = player or self.agent_selection

    player_idx: int = self.agent_name_mapping[player]

    all_moves = product(range(8), repeat=2)
    all_valid_moves = [move for move in all_moves if self.check_action_valid(move, player_idx)]
    return None if len(all_valid_moves) < 1 else random.choice(all_valid_moves)



# import pytest

# import gym
from qwertyenv.ensure_valid_action_pz import EnsureValidAction

from tianshou.data import Collector
from tianshou.env import DummyVectorEnv, PettingZooEnv
from tianshou.policy import MultiAgentPolicyManager, RandomPolicy


from qwertyenv.collect_coins_pz import CollectCoinsEnv

from pettingzoo.test import api_test


def test_evaluate():

  env = CollectCoinsEnv(pieces=['rock', 'rock'])

  action = None

  def another_action_taken(action_taken):
    nonlocal action
    action = action_taken

  # Wrapping the original environment as to make sure a valid action will be taken.
  env = EnsureValidAction(
      env,
      env.check_action_valid,
      env.provide_alternative_valid_action,
      another_action_taken
  )

  api_test(env, num_cycles=10, verbose_progress=False)


def test_with_tianshou():

  action = None

  # env =  gym.make('qwertyenv/CollectCoins-v0', pieces=['rock', 'rock'])

  env = CollectCoinsEnv(pieces=['rock', 'rock'], with_mask=True)

  def another_action_taken(action_taken):
    nonlocal action
    action = action_taken

  # Wrapping the original environment as to make sure a valid action will be taken.
  env = EnsureValidAction(
      env,
      env.check_action_valid,
      env.provide_alternative_valid_action,
      another_action_taken
  )

  env = PettingZooEnv(env)

  policies = MultiAgentPolicyManager([RandomPolicy(), RandomPolicy()], env)

  env = DummyVectorEnv([lambda: env])

  collector = Collector(policies, env)

  result = collector.collect(n_step=200, render=0.1)



from itertools import product
import gymnasium as gym
import random

from qwertyenv.collect_coins_game import CollectCoinsGame


class CollectCoinsEnv(gym.Env):
  """
  CollectCoins Gym environment.

  This a Chess like, two players, turns, board game.
  The aim of the game is to collect more coins than the opponent.
  On each turn a the player's piece makes a move.
  If a coin is present in the destination the count is increased for the player.
  The game ends when there are not more coins on the board to collect. We then compare which player got the most.
  """

  def __init__(self, pieces=['rock', 'rock'], player=0):
    """
    player 0 is the first entry, AKA white
    player 1 is the second entry, AKA black
    """

    obs_space = dict(
      board = gym.spaces.Box(low=0, high=1, shape=(8 * 8,), dtype=bool),
      player = gym.spaces.MultiDiscrete([8, 8]),
      other_player = gym.spaces.MultiDiscrete([8, 8]),
    )    
    self.observation_space = gym.spaces.Dict(spaces=obs_space)
    self.action_space = gym.spaces.MultiDiscrete([8, 8])

    self.pieces = pieces
    self.player = player
    self.other_player = 1 - player
    self.game = None
    self.previous_coins = None
    self.reset()

  def reset(self, seed=None, options=None):
    self.game = CollectCoinsGame(self.pieces)
    self.previous_coins = 0
    return self._get_observation(), {} 

  def step(self, action):
    self.game.make_move(self.player, action)
    if not self.game.is_done():
      self._play_other()
    info = {}
    return self._get_observation(), self._calc_reward(), self.game.is_done(), False, info

  def _play_other(self):
    """
    a random move for now
    """

    move = self.provide_alternative_valid_action(None, self.other_player)
    self.game.make_move(self.other_player, move)

  def _get_observation(self):
    return dict(
      board=self.game.board.flatten(),
      player=self.game.locations[self.player],
      other_player=self.game.locations[self.other_player]
    )

  def _calc_reward(self):
    current_coins = self.game.coins[self.player]
    previous_coins = self.previous_coins
    self.previous_coins = current_coins
    done = self.game.is_done()
    if done:
      draw = self.game.coins[self.player] == self.game.coins[self.other_player]
      win = self.game.coins[self.player] > self.game.coins[self.other_player]
      return 0 if draw else (1 if win else -1)
    else:
      return (current_coins - previous_coins) * 0.01

  def render(self, *args, **argv):
    self.game.render()

  def check_action_valid(self, action, player=None) -> bool:
    """
    This helper function can be used when initializing EnsureValidAction wrapper as an example (see in examples).
    """

    return self.game.valid_move(
      (self.player if player is None else player),
      action
    )

  def provide_alternative_valid_action(self, action, player=None):
    """
    This helper function can be used when initializing EnsureValidAction wrapper as an example (see in examples).

    Select a random valid move.
    TOOD: maybe consult the provided action(move) and provide an action that is as close as possible.
    """

    all_moves = product(range(8), repeat=2)
    all_valid_moves = [move for move in all_moves if self.check_action_valid(move, player)]
    return None if len(all_valid_moves) < 1 else random.choice(all_valid_moves)



import pytest

import gym
from qwertyenv import UpDownLeftRight


def test_up_down_left_right():
  env =  gym.make('qwertyenv/CollectCoins-v0', pieces=['rock', 'rock'])

  env = UpDownLeftRight(env, lambda: (5, 5))
  assert env.action(0) == (5, 4)
  assert env.action(1) == (5, 6)
  assert env.action(2) == (4, 5)
  assert env.action(3) == (6, 5)



from qwertyenv.collect_coins_pz import CollectCoinsEnv
from qwertyenv.pz_to_gymnasium_wrapper import PZ2GymnasiumWrapper
from qwertyenv.ensure_valid_action_pz import EnsureValidAction
# import gym
import qwertyenv
import numpy as np

import torch

import tianshou
from tianshou.data import Collector
from tianshou.env import DummyVectorEnv


action = None

# env =  gym.make('qwertyenv/CollectCoins-v0', pieces=['rock', 'rock'])

pz_env = CollectCoinsEnv(pieces=['rock', 'rock'], with_mask=True)

def another_action_taken(action_taken):
    global action
    action = action_taken

# Wrapping the original environment as to make sure a valid action will be taken.
pz_env = EnsureValidAction(
  pz_env,
  pz_env.check_action_valid,
  pz_env.provide_alternative_valid_action,
  another_action_taken
)

def act_player_1(obs):
    return (0, 0)

gym_env = PZ2GymnasiumWrapper(pz_env, act_others={'player_1': act_player_1})

policy = tianshou.policy.RandomPolicy()

def play(num_episodes: int = 1):

  seed = 1
    
  # ======== environment setup =========
  train_envs = DummyVectorEnv([lambda: gym_env])
  test_envs = DummyVectorEnv([lambda: gym_env])
  # seed
  np.random.seed(seed)
  torch.manual_seed(seed)
  train_envs.seed(seed)
  test_envs.seed(seed)
    
  rewards = []

  policy.eval()
  # pdb.set_trace()
  test_collector = Collector(policy, test_envs, exploration_noise=False)



  result = test_collector.collect(n_step=100000, render=False) # n_episode=num_episodes,
    
  print(result)
  rewards.extend(result['rews'])    

  print(np.mean(rewards))


if __name__ == "__main__":
   play(1)



from abc import ABC, abstractmethod
from typing import List, Protocol

import numpy as np


class Game(Protocol):
  ...


class Piece(ABC):
  """
  Either a rock or a knight (Chess like).
  """
  def __init__(self, game: Game, player: int) -> None:
    self.game = game
    self.player = player

  @abstractmethod
  def __repr__(self) -> str:
    pass
  
  @abstractmethod
  def valid_move(self, move) -> bool:
    pass


class Rock(Piece):
  """
  Rock
  """
  def __init__(self, game: Game, player: int) -> None:
    super().__init__(game, player)

  def __repr__(self) -> str:
    return f'{"w" if self.player == 0 else "b"}R'
  
  def valid_move(self, move) -> bool:
    current_location = self.game.locations[self.player]
    abs_move_row = abs(move[0] - current_location[0]) 
    abs_move_col = abs(move[1] - current_location[1]) 
    return (
      ((abs_move_row == 1) and (abs_move_col == 0))
      or
      ((abs_move_row == 0) and (abs_move_col == 1))
    )


class Knight(Piece):
  """
  Knight
  """
  def __init__(self, game: Game, player: int) -> None:
    super().__init__(game, player)

  def __repr__(self) -> str:
    return f'{"w" if self.player == 0 else "b"}K'
  
  def valid_move(self, move) -> bool:
    current_location = self.game.locations[self.player]
    abs_move_row = abs(move[0] - current_location[0]) 
    abs_move_col = abs(move[1] - current_location[1]) 
    return (
      ((abs_move_row == 1) and (abs_move_col == 2))
      or
      ((abs_move_row == 2) and (abs_move_col == 1))
    )


class CollectCoinsGame:
  """
  CollectCoinsGame
  """

  def __init__(self, pieces=['rock', 'rock']):
    assert len(pieces) == 2
    assert all(piece in ['rock', 'knight'] for piece in pieces)
    game = self
    self.pieces: List[Piece] = [
      Rock(game, i) if piece == 'rock' else Knight(game, i) for i, piece in enumerate(pieces)
    ]
    self.board = np.full((8, 8), True, dtype=bool)
    self.locations = [(0, 0), (7, 7)]
    for loc in self.locations:
      self.board[loc] = False
    self.coins = [0, 0]
    self.turn = 0

  def make_move(self, player, move) -> None:
    move = tuple(move)
    assert self.turn == player
    if move is not None:
      assert self.valid_move(player, move)
      self.coins[player] += int(self.board[move])
      self.board[move] = False
      self.locations[player] = move
    self.turn = 1 - self.turn

  def valid_move(self, player, move):
    move = tuple(move)
    if any(c < 0 or c > 7 for c in move):
      return False
    if any(location == move for location in self.locations):
      return False
    return self.pieces[player].valid_move(move)

  def render(self):
    horizontal = '-' * (self.board.shape[1] * 4 + 1)
    print(horizontal)
    for which_row, row in enumerate(self.board):
      row_copy = [f'{" "}{x}' for x in map({0: ' ', 1: '$'}.get, row)]
      for i, (player_location, player_piece) in enumerate(zip(self.locations, self.pieces)):
        if player_location[0] == which_row:
          row_copy[player_location[1]] = str(player_piece)
      print("|" + " |".join(row_copy) + " |")
      print(horizontal)
    print()
    print(f'{self.coins[0]}/{self.coins[1]}')

  def is_done(self) -> bool:
    return np.sum(self.board) < 1



from typing import TypeVar, Callable
import gymnasium as gym
from .action_wrapper_pz import ActionWrapper


Action = TypeVar("Action")


class EnsureValidAction(ActionWrapper):
  """
  A gym environment wrapper to help with the case that the agent wants to take invalid actions.
  For example consider a Chess game, where you let the action_space be any piece moving to any square on the board,
  but then when a wrong move is taken, instead of returing a big negative reward, you just take another action,
  this time a valid one. To make sure the learning algorithm is aware of the action taken, a callback should be provided.
  """
  def __init__(self, env: gym.Env,
    check_action_valid: Callable[[Action], bool],
    provide_alternative_valid_action: Callable[[Action], Action],
    alternative_action_cb: Callable[[Action], None]):

    super().__init__(env)
    self.check_action_valid = check_action_valid
    self.provide_alternative_valid_action = provide_alternative_valid_action
    self.alternative_action_cb = alternative_action_cb

  def action(self, action: Action) -> Action:
    if self.check_action_valid(action):
      return action
    alternative_action = self.provide_alternative_valid_action(action)
    self.alternative_action_cb(alternative_action)
    return alternative_action
  


from pettingzoo.utils.wrappers import BaseWrapper
import gymnasium as gym


class ActionWrapper(BaseWrapper):
  def __init__(self, env: gym.Env):
    super().__init__(env)

  # def reset(self, seed=None, return_info=True, options=None):
  #   self.env.reset(seed=seed, return_info=return_info, options=options)

  def step(self, action):
    action = self.action(action)
    self.env.step(action)

  def action(self, action):
    pass

  def render(self, *args, **kwargs):
    self.env.render(*args, **kwargs)

  @property
  def agent_selection(self):
    return self.env.agent_selection
  
  @agent_selection.setter
  def agent_selection(self, agent):
    self.env.agent_selection = agent


import gymnasium as gym
from typing import Tuple, Callable
import numpy as np


class UpDownLeftRight(gym.ActionWrapper):
  """
  A gym environment wrapper to enable U/D/L/R action space when the wrapped environment
  actually expects a destination in cartesian coordinates.
  """

  def __init__(self, env: gym.Env,
               get_current_location: Callable[[], Tuple[int, int]]):

    super().__init__(env)
    self.get_current_location = get_current_location

    self.action_space = gym.spaces.Discrete(4)

    self.udlr_action_to_board_action = {
      0: (0, -1), # up
      1: (0, +1), # down
      2: (-1, 0), # left
      3: (+1, 0), # right
    }

  def action(self, action: int) -> Tuple[int, int]:
    current_location = self.get_current_location()
    return tuple(np.add(current_location, self.udlr_action_to_board_action[action]))
